// code compiled and tested with Arduino 1.8.7 & Teensy Loader 1.44

// v1.65 4/26/2019 removed duplicate chipSelect for SD logging.
// v1.64 4/21/2019 removed some remnant debug lines.
// v1.63 3/24/2019 discard CRC failed messages, increase teensy serial 3 RX buffer size, modify serial1.c: #define SERIAL1_RX_BUFFER_SIZE     4048
                   // serial1.c locations
                   //on Mac: /Applications/Arduino.app/Contents/Java/hardware/teensy/avr/cores/teensy3/serial1.c
                   // on Windows: C:\Program Files (x86)\Arduino\hardware\teensy\avr\cores\teensy3\serial1.c
                   // changed isTrackFileID function to be based on target identity (AddressType) instead of address code
// v1.62 2/10/2019 added missing health bytes in the GDLcontrol protocol output
// v1.61 11/15/2018 changed isTrackFileID back to a specific address range
// v1.60 11/15/2018 moved configurations to include files, fixed intruder N number output
// v1.59 11/14/2018 fixed serial.print(x,BYTE) issue that was deprecated in Teensyduino 1.44/ Arduino 1.8.7. Using serial.write() instead.
// v1.58 9/2/2018  fixed SDCARD define, it would throw warnings when commented out, same with ALERTS. dumb monkey, no bananas :)
// v1.57 8/26/2018  added ALERTS define to be able to disable calculated alerts
// v1.56 8/1/2018  added SD card logging for in/out data GDL90 data streams.
// v1.55 7/26/2018  added Shadin Z format altitude input (enable  define SHADINZ)
// v1.54 7/16/2018  changed isTrackFileID function to be based on target identity (AddressType) instead of address code. t=3 TIS-B with track file ID will be treated as trackFileID for ghost filtering
// v1.53 5/24/2018   removed SD card code and includes
// v1.52 5/6/2018    added positive closure max time 
// v1.51 5/6/2018    positive closure + distance & alt
                      //friend blunting/position accuracy degraded to 1;
                      //better relative altitude filter, fuzzy altitude check for projected relAlt.
                      //hold alert for x seconds
// v1.50  05/4/2018   changed SecondsToClose calculations, added velocity vector intersection checking, changed altitude alert filter to calculate at secondstoclose, added minimum distance alert for catching close range parallel traffic.
// v1.44  04/26/2018   filter ownModeSCode, just in case Echo would send it in the traffic stream
// v1.43  04/26/2018   skip coasting if it's a trackfile id and it's within ghost filter distance
//                     do not ghostfilter friends
// v1.42 - 04/25/2018 added vector based calculation for alerts, and time projected altitude closure.
//                    added friendly array and filter (no alerts on friends, degrade targets on friends to display blunt nosed when [FRIEND_dist_start] miles out)
//                    friends filter can be turned of with FRIENDS directive
// v1.41 - 04/19/2018  added custom Traffic Alert settings: alert_alt_diff, alert_dist_diff
// v1.40 - 04/08/2018  escaping double 0x10s in the remote GTX protocol

#define _GNU_SOURCE // needed for strcasestr string function
#include <stdint.h>
#include <Arduino.h>
#include "Aircraft_Settings.h"

//#define DEBUG

//setup timers
IntervalTimer bytesOutTimer;


long int serialDelay=1000000/outputBufferSize; 
int intruderCount=0;
const long DEGREE = 600000; // GPS routines work in units of 1e-4 arc minutes. Thus one degree is represented as 600,000 units:
uint16_t Crc16Table[256];
uint16_t receive_buffer[2048];
int inputBufferSize=0;
byte message[2048];

// output buffers
byte tempBuffer[outputBufferSize]; // for building the buffer
int tempBufferFillIndex=0;

byte outputBuffer[outputBufferSize]; // for sending the buffer
int outputBufferLength=0;
 
byte traffic[27];
int messagesize=0;
uint16_t messageCrc;
uint16_t computedCrc;
bool escapeChar=false;
bool escapedChar=false;
int outputIndex=0;
bool outputFlow=false;
uint32_t bytesSent=0;
uint32_t lastbytesSent=0;
bool remoteEscape=false;

// define intruder type
typedef struct
 {
     uint8_t Alert=0x00;
     byte AddressType;
     int32_t Address;
     double Lat;
     double Lon;
     int32_t Alt;
     uint8_t Misc;
     int8_t Nic;
     int8_t NacP;
     uint16_t Speed;
     int16_t VertSpeed;
     float Track;
     int8_t EmitterCategory;
     char Name[8]; // callsign    
     int8_t priorityCodeSpare;
     unsigned long LastUpdate;
     unsigned long coastUpdate;   
     int16_t geoAlt;
     int32_t RelAlt;
     float Dist;
     float Bearing;
     bool Coasted=0;     
     int16_t BaroDiff=0;
     unsigned long AlertTime;
 }  intruder;
// create array of intruders 
intruder intruders[maxIntruders];
intruder intrudersOut[maxIntrudersOut];
intruder ownShip;
intruder oneIntruder;
typedef struct
  {
  byte message[438];
  } uplinkMessage;  
uplinkMessage uplinkMessageList[maxUplinkMessages];

uint8_t uplinkMessageCount=0;
byte ownshipMessage[28];
byte heartbeatMessage[7];
byte geoaltMessage[5];
uint8_t totalIntruders=0;
uint8_t totalIntrudersOut=0;
unsigned long lastTrafficUpdate=millis();
unsigned long timeElapsed=0;
unsigned long heartbeatTime=millis();
typedef struct
 {  
     int32_t Address; // 0 if entry doesn't exist
     char Nnumber[8]; // callsign
     char Manufacturer[8];
     char Model[8];
     unsigned long LastAccessTime;
     bool valid;  // true if it has useable data
 }  planeInfo;

// store up to 200 planeinfo
const int maxPlaneInfoElements=200;
planeInfo planeInfoArray[maxPlaneInfoElements];

int callsignType=0; // cycle between calssign display types

int icarusBufferSize=0;
char icarusAlt[5];
char icarus_buffer[10];

int remoteBufferSize=0;
byte remote_buffer[16];
bool remoteIdent=0;
byte remoteMode=0;
uint16_t remoteSquawk=1200;


// SD card includes and variables
#ifdef SDCARD
#include <SD.h> 
#include <SPI.h> 
const int chipSelect = BUILTIN_SDCARD;
File inLogFile;
File outLogFile;
File ListFile;
File PlaneInfoFile;

char inLogBuffer[80000];
char outLogBuffer[80000];
int inLogBufferIndex=0;
int outLogBufferIndex=0;
char filenameInLog[14];
char filenameOutLog[14];
bool sdLogging=false;
char serialCmdChar;             // usb serial command character
int serialCmdBufferSize = 0;        // usb serial command buffer size
char serialCmdBuffer[50];       // usb serial command buffer
char listfileFileName[14];      // file name for sd card file listing operations
bool listfileIsDirectory; 
#endif



void setup() {
crcInit();

// init SD card

#ifdef SDCARD
if (!SD.begin(chipSelect)) Serial.println("SD card initialization failed! Logging not available.");
#endif


#ifdef SDLOGGING
sprintf(filenameInLog,"inlog.log");
sprintf(filenameOutLog,"outlog.log");
int fileCount=0;
while (SD.exists(filenameInLog) || SD.exists(filenameOutLog))
      {
      fileCount++;  
      sprintf(filenameInLog,"inlog%d.log",fileCount);
      sprintf(filenameOutLog,"outlog%d.log",fileCount);
      }
Serial.print("inLog file:"); Serial.println(filenameInLog);
Serial.print("outLog file:"); Serial.println(filenameOutLog);
Serial.println();
sdLogging=true;
Serial.println("SD card initialized");
inLogFile = SD.open(filenameInLog, FILE_WRITE);
if (inLogFile) {
             inLogFile.close();
             } else 
                  { 
                  Serial.print("inLogFile opening error");
                  sdLogging=false;
                  }
outLogFile = SD.open(filenameOutLog, FILE_WRITE);
if (outLogFile) {
             outLogFile.close();
             } else
                  {                              
                  Serial.print("outLogFile opening error");
                  sdLogging=false;
                  }                                      
#endif // sdcard



// initialize serial port
Serial.begin(115200); // USB port for debug
Serial1.begin(gdl90_inbaud); // receive GDL90 via RS232
Serial2.begin(gdl90_outbaud); // send filtered GDL90
Serial3.begin(icarus_inbaud); // receive icarus data
Serial4.begin(gtxremote_inbaud); // receive GTX327 or GTX330 data

// for GDLCONTROL setup: serial1= GDL90-IN, serial2=GDL90filtered-OUT, serial3=ICARUS-IN/GDL-OUT, serial4=GTX-REMOTE-IN

//[0x7E 0x00 0x81 0x41 0xDB 0xD0 0x08 0x02 0xB3 0x8B 0x7E] , example message from doc
//byte message[] = {0x00, 0x81, 0x41, 0xDB, 0xD0, 0x08, 0x02};

//uint16_t crc=0;
delay(2000); //wait for things to stabilize
Serial.println("GDL90filter Boot");

inputBufferSize=0;

bytesOutTimer.begin(sendGdl90Bytes, serialDelay);
}

void loop() {

  // read from port 1
  if (Serial1.available())
  {
  int inByte = Serial1.read();
 
  if (inputBufferSize==0)
    {
    //empty buffer, look for flag byte
    if (inByte==0x7E)
          {
          receive_buffer[inputBufferSize]=inByte; // else disregard character until we find the start of a message
          inputBufferSize++;
          }
          #ifdef DEBUG
          else Serial.print("@");
          #endif // debug
    }
    else
        {             
         if (!(inByte==0x7E && inputBufferSize==1))
            {
            if (inByte==0x7D)
              {
              escapeChar=true;
              } else
                    {                    
                    if (escapeChar)
                      {
                      receive_buffer[inputBufferSize]=inByte ^ 0x20;
                      escapeChar=false;
                      escapedChar=true;
                      }
                      else 
                        {
                        receive_buffer[inputBufferSize]=inByte; // add character to buffer  
                        escapedChar=false;
                        }
                    inputBufferSize++;
                    }
            }
         }            
              
            // check for end flag
            if (inputBufferSize>3 && receive_buffer[inputBufferSize-1]==0x7E && !escapedChar) // closing flag 7e
              {
              //process message;
              // sd logging inout
              #ifdef SDCARD
                      if (sdLogging && inLogBufferIndex<80000)
                        {
                        // add timestamp
                        char timestamp[12];
                        sprintf (timestamp, "%012ld:", millis());
                        memcpy(inLogBuffer+inLogBufferIndex,timestamp,13);
                        inLogBufferIndex+=13;
                        for (int i=0;i<=inputBufferSize-1;i++)
                            {
                            char hexChar[3];
                            sprintf(hexChar,"%02X ",receive_buffer[i]);
                            memcpy(inLogBuffer+inLogBufferIndex,hexChar,3);
                            inLogBufferIndex+=3;
                            }
                        // addnew line
                         inLogBuffer[inLogBufferIndex]=0x0A;
                         inLogBufferIndex++;                            
                        }                             
              #endif
              
              //$message=array_slice($buffer, 1, count($buffer)-4);
               messagesize=0;
              // qualify message type
               uint16_t messageID=receive_buffer[1];                            
              #ifdef DEBUG 
              Serial.println();
              #endif
              if (messageID!=1)
                  {
                  // use message
                  for (int i=1;i<inputBufferSize-3;i++)
                  {
                  message[messagesize]=receive_buffer[i];
                  messagesize++;
                  }               
                  messageCrc=receive_buffer[inputBufferSize-3]+256*receive_buffer[inputBufferSize-2];
                  computedCrc=crcCompute(message,messagesize);
                  // print message
                  if (messageCrc==computedCrc)
                  {
                  
                   switch (message[0]) {
                                      case 0: 
                                            {
                                            heartbeatTime=millis(); // update heartbeat time
                                            #ifdef DEBUG
                                            Serial.print(millis());Serial.println(": HEARTBEAT -> ");                                            
                                            #endif                                           
                                            copyMessage(message,heartbeatMessage,messagesize); // grab heartbeat message
                                            //printMessage(message,messagesize,"SavedHEARTBEAT");                                            
                                            //Serial.println("processmessages");
                                            prepareTrafficOut();
                                            uplinkMessageCount=0;                                            
                                            break;
                                            }
                                            
                                      //case 2: Serial.print(millis());Serial.print(": INIT -> ");break;
                                      //case 9: Serial.print(millis());Serial.print(": HEIGHT -> ");break;
                                      case 10:
                                          {
                                          intruderCount=0;
                                          #ifdef DEBUG
                                          Serial.print(millis());Serial.print(": OWNSHIP -> ");
                                           for (int k=19;k<27;k++)
                                              {
                                               Serial.write(message[k]);
                                               ownShip.Name[k-19]=message[k];
                                              }
                                           Serial.print(" ");
                                          #endif     
                                          // decode and update ownship                                                                                    
                                          //byte trafficAlert=message[1]>>4; // shift right 4 times to get first 4 bits                                                             
                                            // Genius post #13: http://forum.arduino.cc/index.php?topic=321296.0
                                            long lat=(((long)(signed char)message[5]) <<16) | ((long)message[6]<<8) | ((long)message[7]);                                        
                                            long lon=(((long)(signed char)message[8]) <<16) | ((long)message[9]<<8) | ((long)message[10]);
                                            ownShip.Lat=lat*180/pow(2,23);
                                            ownShip.Lon=lon*180/pow(2,23);                                        
                                            //ownShip.Alt=(16*message[11]+(message[12]>>4))*25-1000;
                                            ownShip.Alt=(message[11]<<4 | (message[12]>>4))*25-1000;
                                            ownShip.Track=message[17]*1.3671875; //360/256
                                            ownShip.Misc=message[12] & B00001111;
                                            ownShip.Speed=(message[14]<<4 | (message[15]>>4));
                                            ownShip.Nic=message[13]>>4;
                                            ownShip.NacP=message[13]&B00001111;
                                            
                                            if (ownShip.Alt!=0 && ownShip.geoAlt!=0) ownShip.BaroDiff=ownShip.Alt-ownShip.geoAlt;
                                            if (ownShip.Lat!=0 && ownShip.Lon!=0) ownShip.LastUpdate=millis();                                          
                                            #ifdef DEBUG
                                            Serial.println();
                                            Serial.write("Ownship Lat: ");
                                            printDouble(ownShip.Lat,100000);
                                            Serial.write(", Long: ");                                      
                                            printDouble(ownShip.Lon,100000);
                                            Serial.write(", Alt: ");
                                            Serial.print(ownShip.Alt);
                                            Serial.write(", Track: ");                                            
                                            Serial.print(ownShip.Track);
                                            Serial.write(", Speed: ");                                            
                                            Serial.print(ownShip.Speed);
                                            Serial.write(", NIC:");Serial.print(ownShip.Nic);Serial.write(", NACp: "); Serial.print(ownShip.NacP);                                            
                                            Serial.write(", Misc: ");                                            
                                            Serial.print(ownShip.Misc);                                            
                                            Serial.println();
                                            #endif // debug
                                            //printMessage(message,messagesize,"SavedOWNSHIP");
                                            copyMessage(message,ownshipMessage,messagesize); // grab ownship message                                                                                                                                                                                                                                    
                                           break; 
                                          } 
                                      case 11:
                                            {
                                            //printMessage(message,messagesize,"SavedGEOALT");  
                                            copyMessage(message,geoaltMessage,messagesize); // grab geoalt message                                           
                                            break;
                                            }
                                      case 20:
                                        {
                                        intruderCount++;                                        
                                        oneIntruder.LastUpdate=millis();
                                        #ifdef DEBUG
                                        Serial.print(millis());Serial.print(": TRAFFIC -> ");
                                        for (int k=19;k<27;k++) Serial.write(message[k]);                                          
                                        Serial.println();
                                        #endif                                        
                                        // decode traffic
                                        for (int k=19;k<27;k++) oneIntruder.Name[k-19]=message[k];                                                                                                                                                                                                                                                                                                                                         
                                        oneIntruder.coastUpdate=oneIntruder.LastUpdate;
                                        //oneIntruder.Alert=message[1]>>4; // shift right 4 times to get first 4 bits
                                        oneIntruder.AddressType=message[1] & B00001111;
                                        oneIntruder.VertSpeed=((message[15]&B00001111)<<8 | message[16])*64;
                                        long lat=0;
                                        long lon=0;                 
                                        // Genius post #13: http://forum.arduino.cc/index.php?topic=321296.0
                                        lat=(((long)(signed char)message[5]) <<16) | ((long)message[6]<<8) | ((long)message[7]);                                        
                                        lon=(((long)(signed char)message[8]) <<16) | ((long)message[9]<<8) | ((long)message[10]);
                                        oneIntruder.Lat=lat*180/pow(2,23);
                                        oneIntruder.Lon=lon*180/pow(2,23);                                        
                                        oneIntruder.Alt=(message[11]<<4 | (message[12]>>4))*25-1000;
                                        oneIntruder.Speed=(message[14]<<4 | (message[15]>>4));
                                        oneIntruder.Dist=HaverSine(ownShip.Lat,ownShip.Lon,oneIntruder.Lat,oneIntruder.Lon);
                                        oneIntruder.Bearing=Bearing(ownShip.Lat,ownShip.Lon,oneIntruder.Lat,oneIntruder.Lon,ownShip.Track);                                                                                                                       
                                        oneIntruder.Address=message[2]<<16 | message[3] <<8 | message[4];
                                        oneIntruder.Nic=message[13]>>4;
                                        oneIntruder.NacP=message[13]&B00001111;
                                        oneIntruder.Track=message[17]*1.3671875;//360/256
                                        oneIntruder.Misc=message[12] & B00001111;
                                        oneIntruder.EmitterCategory=message[18];
                                        oneIntruder.priorityCodeSpare=message[27];                                        
                                        
                                        //Serial.write("-> ");Serial.print(oneIntruder.Misc,BIN);                                                                                                                 
                                        //  ownShip.BaroDiff=ownShip.Alt-ownShip.geoAlt;
                                        // fix baro altitude if it's not received
                                        if (ownShip.Alt==0) ownShip.Alt=ownShip.geoAlt;                                          
                                        oneIntruder.RelAlt=oneIntruder.Alt-ownShip.Alt;
                                        oneIntruder.Coasted=0;                                      
                                        //printMessage(message,messagesize,"SavedTRAFFIC");

                                        // filter by altitude and distance
                                        if (oneIntruder.RelAlt <=filterAltitudeAbove
                                        && oneIntruder.RelAlt >= -filterAltitudeBelow  && oneIntruder.Dist <= filterDistance && ghostFilter(oneIntruder) && oneIntruder.Address!=ownModeSCode) AddIntruder(oneIntruder);                                          
                                       
                                        break;   
                                        }
                                      case 7:
                                        if (uplinkMessageCount< maxUplinkMessages)
                                            {
                                            copyMessage(message,uplinkMessageList[uplinkMessageCount].message,messagesize); // grab uplink message                                                                               
                                            //printMessage(uplinkMessageList[uplinkMessageCount].message,messagesize,"SavedUPLINK");                                        
                                            uplinkMessageCount++;
                                            }
                                        
                                        break;                      
                                      } // switch
                      } // if CRC


                      
                      #ifdef DEBUG
                       else {
                            Serial.print(" [CRC FAIL] ");
                            Serial.print(messageCrc,HEX);Serial.print("<>");Serial.print(computedCrc,HEX);Serial.println();
                            };
                     #endif
                } // if messageid!=1                   
                                     
               
               inputBufferSize=0;  // reset message buffer
                       
              } // end message processing


  } // if serial1.available

#ifdef GDLCONTROL

#ifdef ICARUS

if (Serial3.available())
  {
// read icarus
//ALT<sp>12345<CR>
char icarusChar = Serial3.read();
Serial.print(icarusChar);
  if (icarusBufferSize==0)
    {    
    //empty buffer, look for ascii A    
    if (icarusChar=='A')
          {
          icarus_buffer[icarusBufferSize]=icarusChar; 
          icarusBufferSize++;
          if (icarusBufferSize>10) icarusBufferSize=0; // check for overflow
          }// else Serial.print("#"); // disregard input until we get a proper starting character (A)
    } else
          {
          if (icarusChar!=char(0x0D)) // not the closing character
              {
              icarus_buffer[icarusBufferSize]=icarusChar;
              icarusBufferSize++;
              if (icarusBufferSize>10) icarusBufferSize=0; // check for overflow
              } else 
                    {
                    // end of message block process message
                    //for (int k=0;k<icarusBufferSize;k++) Serial.print(icarus_buffer[k]);
                    
                    if (icarus_buffer[0]=='A' && icarus_buffer[1]=='L' && icarus_buffer[2]=='T')
                      {
                       #ifdef DEBUG
                       Serial.println();
                       Serial.print("Icarus: "); 
                       #endif
                       for (int i=4;i<=8;i++)
                            {
                            icarusAlt[i-4]=icarus_buffer[i];
                            #ifdef DEBUG
                            Serial.print(icarus_buffer[i]);
                            #endif
                            }
                       #ifdef DEBUG
                       Serial.print(" > ");
                       #endif
                      sendGDLAltitude();
                      }
                    // reset buffer
                    icarusBufferSize=0;
                    }
              
          }
  } // if serial.available

 #endif

#ifdef SHADINZ

if (Serial3.available())
  {
// read Shadin Z format
//ZD+12345<CR>
char icarusChar = Serial3.read();
//Serial.print(icarusChar);
  if (icarusBufferSize==0)
    {    
    //empty buffer, look for ascii A    
    if (icarusChar=='Z')
          {
          icarus_buffer[icarusBufferSize]=icarusChar; 
          icarusBufferSize++;
          if (icarusBufferSize>10) icarusBufferSize=0; // check for overflow
          } else Serial.print("&"); // disregard input until we get a proper starting character (D)
    } else
          {
          if (icarusChar!=char(0x0D)) // not the closing character
              {
              icarus_buffer[icarusBufferSize]=icarusChar;
              icarusBufferSize++;
              if (icarusBufferSize>10) icarusBufferSize=0; // check for overflow
              } else 
                    {
                    // end of message block process message
                    //for (int k=0;k<icarusBufferSize;k++) Serial.print(icarus_buffer[k]);
                    
                    if (icarus_buffer[0]=='Z' && icarus_buffer[1]=='D' && (icarus_buffer[2]=='+' || icarus_buffer[2]=='-'))
                      {
                       #ifdef DEBUG
                       Serial.println();
                       Serial.print("ShadinZ: "); 
                       #endif
                       for (int i=3;i<=6;i++)
                            {
                            icarusAlt[i-3]=icarus_buffer[i];
                            #ifdef DEBUG
                            Serial.print(icarus_buffer[i]);
                            #endif
                            }
                       // add a zero at the end, 10s of feet in ShadinZ
                       icarusAlt[4]='0';
                       #ifdef DEBUG
                       Serial.print("0 > ");
                       #endif
                      sendGDLAltitude();
                      }
                    // reset buffer
                    icarusBufferSize=0;
                    }
              
          }
  } // if serial3.available

 #endif  

if (Serial4.available())
  {
byte remoteByte= Serial4.read();
   if (remoteBufferSize==0)
      {
      remoteEscape=false; // init 0x10 escape flag
      // empty buffer look for 0x10;
      if (remoteByte==0x10)
          {
          remote_buffer[remoteBufferSize]=remoteByte;
          remoteBufferSize++; 
          } else Serial.print("*");       
      } else
           if (remoteBufferSize==1)
              {
              if (remoteByte==0x36)
                  {
                  remote_buffer[remoteBufferSize]=remoteByte;
                  remoteBufferSize++;    
                  } else {
                          Serial.print("*");
                          remoteBufferSize=0; 
                         }                  
              } else
                    {
                    // got past the starting point
                    remote_buffer[remoteBufferSize]=remoteByte;
                    remoteBufferSize++;
                    // check for end
                    if (remote_buffer[remoteBufferSize-1]==0x03 && remote_buffer[remoteBufferSize-2]==0x10)
                        {
                        // process remote message                        
                         #ifdef DEBUG
                        Serial.println();
                        Serial.print("GTX Remote: "); 
                        for (int i=0;i<remoteBufferSize;i++)
                            {                            
                            Serial.print(remote_buffer[i],HEX);
                            Serial.print(" ");                                                        
                            }                            
                        Serial.println();
                        #endif
                        if(remote_buffer[13]==remoteCRC(remote_buffer))
                            {
                            #ifdef DEBUG
                            Serial.print (" CRC OK ");
                            #endif
                            // get data from message block
                            if (remote_buffer[3]==0x00) remoteIdent=false; else remoteIdent=true;
                            remoteMode=remote_buffer[6];
                            if (remote_buffer[7]==0x00) remoteSquawk=remote_buffer[9]<<8 | remote_buffer[8]; else remoteSquawk=remote_buffer[11]<<8 | remote_buffer[10];
                            #ifdef DEBUG
                            Serial.print(",Ident: "); Serial.print(remoteIdent,BIN);
                            Serial.print(",Mode: "); Serial.print(remoteMode,HEX);
                            Serial.print(",Squawk: "); Serial.println(remoteSquawk,DEC);                           
                            #endif
                            sendGDLMode(); // send it out on serial  
                            }
                          
                        remoteBufferSize=0; // reset buffersize    
                        }
                      //escape 0x10 in the message, if there's two 10s in a row discared the last one.
                      if (remote_buffer[remoteBufferSize-1]==0x10 && remote_buffer[remoteBufferSize-2]==0x10)
                      {
                       if (!remoteEscape)
                          {
                          remoteBufferSize--;
                          remoteEscape=true;
                          } else remoteEscape=false;
                      }
                    } 
   if (remoteBufferSize>16) remoteBufferSize=0; // check for overflow 
  } // if serial4.available

 #endif 

#ifdef SDLOGGING
// look for serial commands
if (Serial.available())
  {
//LIST!
//DELETE filename!
//PRINT filename!
//STOP!
//START!
//FORMAT!
char serialCmdChar = Serial.read();    
  if (serialCmdChar!=char(0x21) && serialCmdChar!=char(0x0D) && serialCmdChar!=char(0x0A) && serialCmdBufferSize<50)
    {    
    serialCmdBuffer[serialCmdBufferSize]=serialCmdChar;
    serialCmdBufferSize++;
    } else
          {
          // process command                    
           if (strcasestr(serialCmdBuffer, "LIST"))
              {
              // list files
              Serial.println("Files list:");
              ListFile=SD.open("/");

              while(true) {
     
                           File entry =  ListFile.openNextFile();
                           if (! entry) {
                                         // no more files                             
                                         break;
                                        }                           
                           
                           if (!entry.isDirectory())
                              {
                              // only list files in root folder, no directories
                              Serial.print(entry.name());
                              Serial.print("\t");
                              Serial.println(entry.size(), DEC);
                              }                             
                           entry.close();
                           }
              
              } else
                     if (strcasestr(serialCmdBuffer, "DELETE"))
                        {
                        // delete file                        
                        memcpy(listfileFileName,serialCmdBuffer+7,serialCmdBufferSize-6);
                        SD.remove(listfileFileName);
                        Serial.println();
                        Serial.print("Deleted: ");
                        Serial.println(listfileFileName);
                        } else
                              if (strcasestr(serialCmdBuffer, "PRINT"))
                                 {
                                 // print file contents
                                 memcpy(listfileFileName,serialCmdBuffer+6,serialCmdBufferSize-5);
                                 ListFile=SD.open(listfileFileName);
                                 if (ListFile) {
                                              Serial.println();
                                              Serial.print(listfileFileName);                                              
                                              Serial.println(":");
                                              // read from the file until there's nothing else in it:
                                              while (ListFile.available()) {
                                               Serial.write(ListFile.read());
                                              }
                                              // close the file:
                                              ListFile.close();
                                            } else {
                                              // if the file didn't open, print an error:
                                              Serial.print("Error opening "); Serial.println(listfileFileName);
                                            }
                                 
                                 } else
                                      if (strcasestr(serialCmdBuffer, "STOP"))
                                         {
                                         // stop SD logging to take file commands
                                         sdLogging=false;
                                         Serial.println("STOPPED SD logging.");
                                         } else
                                              if (strcasestr(serialCmdBuffer, "START"))
                                                 {
                                                 // start SD logging
                                                 sdLogging=true;
                                                 Serial.println("STARTED SD logging.");
                                                 } else
                                                      if (strcasestr(serialCmdBuffer, "FORMAT"))
                                                          {
                                                          // delete file                        
                                                          memcpy(listfileFileName,serialCmdBuffer+7,serialCmdBufferSize-6);
                                                          //Serial.print("DeleteCmd:");Serial.println(listfileFileName);                                                          
                                                          ListFile=SD.open("/");
                                                          while(true) {
                                                   
                                                                         File entry =  ListFile.openNextFile();
                                                                         if (! entry) {
                                                                                       // no more files                             
                                                                                       break;
                                                                                      }                           
                                                                         strcpy(listfileFileName,entry.name());
                                                                         if (entry.isDirectory()) listfileIsDirectory=true; else listfileIsDirectory=false;
                                                                         entry.close();                                                                         
                                                                         if (!listfileIsDirectory) SD.remove(listfileFileName); else SD.rmdir(listfileFileName);                                                                      
                                                                         }
                                                          Serial.println("Deleted all files on SD card.");
                                                          }

                                                                                                                                                                          

          
          // reset cmdBuffer
          memset(serialCmdBuffer,0,sizeof(serialCmdBuffer));
          serialCmdBufferSize=0;    
          }
  } // if serial.available
saveLogBuffer();            
#endif // serial commands
} // loop

byte remoteCRC(byte remoteMessage[])
  {
  byte result=0;  
  for (int i=1;i<=12;i++)
      {
      result=result-remoteMessage[i];
      }
  return result;    
  }

void printMessage(byte message[], int messagesize, String type)
  {
  Serial.print(type+": ");
  for (int i=0;i<messagesize;i++)
      {      
      Serial.print(message[i],HEX);
      Serial.print(" ");
      }
  Serial.println();
  }

void sendGDLAltitude()
  {
  char gdl_buffer[17];
  byte gdlBufferSize=0;
  byte gdlChecksum=0;

  gdl_buffer[gdlBufferSize]='#';gdlBufferSize++;
  gdl_buffer[gdlBufferSize]='A';gdlBufferSize++;
  gdl_buffer[gdlBufferSize]='L';gdlBufferSize++;
  gdl_buffer[gdlBufferSize]=' ';gdlBufferSize++; 
  if (icarusAlt[0]=='-')
                {
                gdl_buffer[gdlBufferSize]='-';gdlBufferSize++; 
                gdl_buffer[gdlBufferSize]='0';gdlBufferSize++;
                for (int i=1;i<5;i++)
                        {
                        gdl_buffer[gdlBufferSize]=icarusAlt[i];gdlBufferSize++;
                        }
                }                
                else
                    {
                     gdl_buffer[gdlBufferSize]='+';gdlBufferSize++; 
                     for (int i=0;i<5;i++)
                        {
                         gdl_buffer[gdlBufferSize]=icarusAlt[i];gdlBufferSize++;
                        } 
                    }
  gdl_buffer[gdlBufferSize]='T';gdlBufferSize++;
  gdl_buffer[gdlBufferSize]='+';gdlBufferSize++;
  gdl_buffer[gdlBufferSize]='1';gdlBufferSize++;
  gdl_buffer[gdlBufferSize]='5';gdlBufferSize++;    
  // send checksum
  for (int i=0;i<gdlBufferSize;i++) gdlChecksum=gdlChecksum+gdl_buffer[i];
  gdl_buffer[gdlBufferSize]=c2h(gdlChecksum>>4);gdlBufferSize++;
  gdl_buffer[gdlBufferSize]=c2h(gdlChecksum);gdlBufferSize++;  
  gdl_buffer[gdlBufferSize]=char(0x0D);gdlBufferSize++; 
  //print to serial
  #ifdef DEBUG
          Serial.print("GDLOUT:");
  #endif          
  for (int i=0;i<gdlBufferSize;i++)
          {          
          Serial3.print(gdl_buffer[i]);
          #ifdef DEBUG
          Serial.print(gdl_buffer[i]);            
          #endif
          }
  #ifdef DEBUG
          Serial.println();
  #endif         
  }


void sendGDLMode()
  {
  char gdl_mode_buffer[17];
  byte gdlModeBufferSize=0;
  byte gdlModeChecksum=0;
  char squawkBuffer[5]; // 4 + 1;
  
  gdl_mode_buffer[gdlModeBufferSize]='^';gdlModeBufferSize++;
  gdl_mode_buffer[gdlModeBufferSize]='M';gdlModeBufferSize++;
  gdl_mode_buffer[gdlModeBufferSize]='D';gdlModeBufferSize++;
  gdl_mode_buffer[gdlModeBufferSize]=' ';gdlModeBufferSize++;
  switch (remoteMode) {
          case 0x00: gdl_mode_buffer[gdlModeBufferSize]='A'; // ON (Mode A)
          break;
          case 0x01: gdl_mode_buffer[gdlModeBufferSize]='C'; // ALT (Mode C)
          break;
          case 0x02: gdl_mode_buffer[gdlModeBufferSize]='O'; // STANDBY
          break;
          default: gdl_mode_buffer[gdlModeBufferSize]='O'; // ground mode? let's make it standby
          }
  gdlModeBufferSize++;        
  gdl_mode_buffer[gdlModeBufferSize]=char(',');gdlModeBufferSize++;
  if (remoteIdent) gdl_mode_buffer[gdlModeBufferSize]=char('I'); else gdl_mode_buffer[gdlModeBufferSize]=char('-');
  gdlModeBufferSize++;
  gdl_mode_buffer[gdlModeBufferSize]=char(',');gdlModeBufferSize++;
  //add squawk code
  sprintf (squawkBuffer, "%04i", remoteSquawk);
  for (int i=0;i<4;i++)
      {
      gdl_mode_buffer[gdlModeBufferSize]=squawkBuffer[i];gdlModeBufferSize++;
      }

   // send health/heartbeat byte
   gdl_mode_buffer[gdlModeBufferSize]=char('0');gdlModeBufferSize++;
   gdl_mode_buffer[gdlModeBufferSize]=char('1');gdlModeBufferSize++;
   
  // send 2 checksum bytes
  for (int i=0;i<gdlModeBufferSize;i++) gdlModeChecksum=gdlModeChecksum+gdl_mode_buffer[i];
  gdl_mode_buffer[gdlModeBufferSize]=c2h(gdlModeChecksum>>4);gdlModeBufferSize++;
  gdl_mode_buffer[gdlModeBufferSize]=c2h(gdlModeChecksum);gdlModeBufferSize++; 
  
  gdl_mode_buffer[gdlModeBufferSize]=char(0x0D);gdlModeBufferSize++;
  //print to serial 
  
  #ifdef DEBUG
          Serial.print("GDLOUT:");
  #endif          
  for (int i=0;i<gdlModeBufferSize;i++)
          {          
          Serial3.print(gdl_mode_buffer[i]);
          #ifdef DEBUG
          Serial.print(gdl_mode_buffer[i]);            
          #endif
          }
  #ifdef DEBUG
          Serial.println();
  #endif  


  
  }

char c2h(char c)
{
 return "0123456789ABCDEF"[0x0F & (unsigned char)c];
}

void copyMessage(byte fromMessage[], byte toMessage[], int messagesize)
{
for (int i=0;i<messagesize;i++)
  {
  toMessage[i]=fromMessage[i];
  }    
}

void crcInit( void )
{
 // initialize crc table 
 uint16_t i, bitctr, crc;
 for (i = 0; i < 256; i++)
 {
 crc = (i << 8);
 for (bitctr = 0; bitctr < 8; bitctr++)
 {
  if (crc & 0x8000) crc = (crc << 1) ^ 0x1021; else crc=(crc << 1) ^ 0x0000;
 }
Crc16Table[i] = crc;
 }
}

uint16_t crcCompute(byte *block,unsigned long int length)
{
 // calculate crc of message block 
 unsigned long int i;
 uint16_t crc = 0;
 for (i = 0; i < length; i++)
 {
 crc = Crc16Table[crc >> 8] ^ (crc << 8) ^ block[i];
 }
 return crc;
}

void printDouble( double val, unsigned int precision){
// prints val with number of decimal places determine by precision
// NOTE: precision is 1 followed by the number of zeros for the desired number of decimial places
// example: printDouble( 3.1415, 100); // prints 3.14 (two decimal places)
//
   Serial.print (int(val));  //prints the int part
   Serial.print("."); // print the decimal point
   unsigned int frac;
   if(val >= 0)
       frac = (val - int(val)) * precision;
   else
       frac = (int(val)- val ) * precision;
   Serial.print(frac,DEC) ;
}

float HaverSine(float lat1, float lon1, float lat2, float lon2)
{
 float ToRad = PI / 180.0;
 float R = 3440.065;   // radius earth in nm
 
 float dLat = (lat2-lat1) * ToRad;
 float dLon = (lon2-lon1) * ToRad; 
 
 float a = sin(dLat/2) * sin(dLat/2) +
       cos(lat1 * ToRad) * cos(lat2 * ToRad) * 
       sin(dLon/2) * sin(dLon/2); 
       
 float c = 2 * atan2(sqrt(a), sqrt(1-a)); 
 
 float d = R * c;
 return d;
}





float Bearing(float lat1, float lon1, float lat2, float lon2, float track)
  {
  float ToRad = PI / 180.0;
  float ToDegrees = 180.0 / PI;
  lat1=lat1 * ToRad;
  lat2=lat2 * ToRad;
  float dLon=(lon2-lon1) * ToRad;
  float y=sin(dLon) * cos(lat2);
  float x=cos(lat1)*sin(lat2)-sin(lat1)*cos(lat2)*cos(dLon);
  float brng=atan2(y,x)*ToDegrees;
  if (brng<=0) brng=360+brng; // wrap around compass value
  brng=brng-track;
  if (brng<-180) brng=brng+360;
  return brng; 
  }

void AddIntruder(intruder NewIntruder)
{
int i;

if (totalIntruders>=maxIntruders)
    {
    // intruder list full, check if distance is closer than the last one, and remove last, and insert this one in the right place.
    if (NewIntruder.Dist < intruders[totalIntruders-1].Dist)
        {
        RemoveIntruder(intruders[totalIntruders-1].Address);
        // made room for new intruder.
        } else return; // new intruder's distance is higher than any of the once in the list, do not add to list
    }

// remove it first then insert, to update location in array based on distance
RemoveIntruder(NewIntruder.Address);

if (totalIntruders>=maxIntruders) return; // don't add it if our buffer is about to overflow

for (i=totalIntruders-1;( i >= 0  && intruders[i].Dist > NewIntruder.Dist);i--)
    {
    intruders[i+1]=intruders[i];  
    }
intruders[i+1]=NewIntruder;
totalIntruders++;
}



int8_t FindIntruder(int32_t Address)
{
 for (int i=0; i<totalIntruders; i++)
    {
    if (intruders[i].Address==Address) return i; 
    }
 return -1;
}

void RemoveIntruder(int32_t Address)
{
int i;
int pos ;
pos=FindIntruder(Address);
if (pos==-1) return; // not found
    for (i=pos; i<totalIntruders-1; i++)
     {        
     intruders[i]=intruders[i+1];
     }
    totalIntruders--;
}

void ExpireIntruders()
{
int i=0;
while (i<totalIntruders)
  {         
     if ((millis()-intruders[i].LastUpdate)>coastTime)
            {
             // expired intruder, remove
             for (int j=i; j<totalIntruders-1; j++)
                 {        
                 intruders[j]=intruders[j+1];
                 }
                 totalIntruders--;
                 // keep i the same, since a new intruder took it's place, we need to check it again.
            }  else i++;
  }

//Serial.println();
//Serial.write("Intruders: ");
//Serial.print(totalIntruders);
//Serial.write(" Times: ");
//for (int k=0;k<totalIntruders;k++)
//  {
//   Serial.print(millis()-intruders[k].LastUpdate);
//   Serial.write(",");
//  }
//Serial.println();

}


void printTraffic()
{
Serial.println();
Serial.write("Total Intruders: ");
Serial.print(totalIntruders);
Serial.println();

for (int i=0; i<totalIntruders;i++)
  {
  Serial.println();
  Serial.print(i);
  Serial.write(": ");
  Serial.print(intruders[i].Address,HEX);
  Serial.write(":");
  Serial.print(intruders[i].AddressType);
  Serial.write(", age: ");
  Serial.print(millis()-intruders[i].LastUpdate);
  Serial.write(", distance: ");
  Serial.print(intruders[i].Dist);
  Serial.write(", speed: ");
  Serial.print(intruders[i].Speed);  
  Serial.write(", track: ");
  Serial.print(intruders[i].Track);
  Serial.write(", alt: ");
  Serial.print(intruders[i].Alt);
  Serial.write(", relalt: ");
  Serial.print(intruders[i].RelAlt);
  Serial.write(", vertspeed: ");
  Serial.print(intruders[i].VertSpeed);
  Serial.write(", coasted: ");
  Serial.print(intruders[i].Coasted);
  Serial.println();
  }
}

void coastTraffic()
{
for (int i=0; i<totalIntruders;i++)
  {
   if ((millis()-intruders[i].coastUpdate) > 1000)
     { 
      if (isTrackFileID(intruders[i].AddressType) && intruders[i].Dist<=ghost_dist_diff) continue; // skip coasting if it's a trackfile id and it's within ghost filter distance
      int updateTime=millis()-intruders[i].coastUpdate;
      float R = 3440.065;
      float ToRad = PI / 180.0;
      float ToDegrees = 180.0 / PI;
      float Lat1=intruders[i].Lat*ToRad;
      float Lon1=intruders[i].Lon*ToRad;
      float Track=intruders[i].Track*ToRad;
      // distance=speed*time in hours;
      float d=intruders[i].Speed*updateTime/(1000.0*3600.0);
      //var φ2 = Math.asin( Math.sin(φ1)*Math.cos(d/R) + Math.cos(φ1)*Math.sin(d/R)*Math.cos(brng));
      float Lat2=asin(sin(Lat1)*cos(d/R)+cos(Lat1)*sin(d/R)*cos(Track));
                          
      //var λ2 = λ1 + Math.atan2(Math.sin(brng)*Math.sin(d/R)*Math.cos(φ1), Math.cos(d/R)-Math.sin(φ1)*Math.sin(φ2));
      float Lon2=Lon1+atan2(sin(Track)*sin(d/R)*cos(Lat1),cos(d/R)-sin(Lat1)*sin(Lat2));
      
      intruders[i].Lat=Lat2*ToDegrees;
      intruders[i].Lon=remainder(Lon2*ToDegrees+540, 360)-180;
        
      intruders[i].Dist=HaverSine(ownShip.Lat,ownShip.Lon,intruders[i].Lat,intruders[i].Lon);
      intruders[i].Bearing=Bearing(ownShip.Lat,ownShip.Lon,intruders[i].Lat,intruders[i].Lon,ownShip.Track);
      //new altitude=altitude + vertrate*time in minutes      
      intruders[i].Alt=intruders[i].Alt+((intruders[i].VertSpeed * updateTime) / 60000.0);
      intruders[i].RelAlt=intruders[i].Alt-ownShip.Alt;
      intruders[i].Coasted=1;
      intruders[i].coastUpdate=millis();
     }
  }
}  

float closingSpeed(intruder myIntruder)
{
 float projSeconds=60.0; // 60 second projection for closign speed calc
 float R = 3440.065;
 float ToRad = PI / 180.0;
 float ToDegrees = 180.0 / PI;
// calculate projected target position in projSeconds;
 float targetLatR=myIntruder.Lat*ToRad;
 float targetLonR=myIntruder.Lon*ToRad;
 float targetTrackR=myIntruder.Track*ToRad;
 float targetDist=myIntruder.Speed*(projSeconds)/3600.0; // the distance target travels in projSeconds
 float projTargetLatR=asin(sin(targetLatR)*cos(targetDist/R)+cos(targetLatR)*sin(targetDist/R)*cos(targetTrackR)); // projected latitude
 float projTargetLonR=targetLonR+atan2(sin(targetTrackR)*sin(targetDist/R)*cos(targetLatR),cos(targetDist/R)-sin(targetLatR)*sin(projTargetLatR)); // projected longitude
 float projTargetLat=projTargetLatR*ToDegrees;
 float projTargetLon=remainder(projTargetLonR*ToDegrees+540, 360)-180;
 //calculate projected Ownship position in projSeconds;
 float ownshipLatR=ownShip.Lat*ToRad;
 float ownshipLonR=ownShip.Lon*ToRad;
 float ownshipTrackR=ownShip.Track*ToRad;
 float ownshipDist=ownShip.Speed*(projSeconds)/3600.0; // the distance target travels in projSeconds
 float projOwnshipLatR=asin(sin(ownshipLatR)*cos(ownshipDist/R)+cos(ownshipLatR)*sin(ownshipDist/R)*cos(ownshipTrackR)); // projected latitude
 float projOwnshipLonR=ownshipLonR+atan2(sin(ownshipTrackR)*sin(ownshipDist/R)*cos(ownshipLatR),cos(ownshipDist/R)-sin(ownshipLatR)*sin(projOwnshipLatR)); // projected longitude
 float projOwnshipLat=projOwnshipLatR*ToDegrees;
 float projOwnshipLon=remainder(projOwnshipLonR*ToDegrees+540, 360)-180;
 // get distance between original and projected positions
 float originalDistance=myIntruder.Dist;
 float projDistance=HaverSine(projTargetLat,projTargetLon,projOwnshipLat,projOwnshipLon);
 // return closing speed
  return (originalDistance-projDistance)*3600.0/projSeconds;
}

bool vectorsIntersect(intruder myIntruder)
{
if (myIntruder.Dist>10) return false; // if target is over 10 miles don't bother calculating
float R = 3440.065;
float ToRad = PI / 180.0;
 float ToDegrees = 180.0 / PI;
// calculate projected target position in projSeconds;
 float targetLatR=myIntruder.Lat*ToRad;
 float targetLonR=myIntruder.Lon*ToRad;
 float targetTrackR=myIntruder.Track*ToRad;
 float targetDist=myIntruder.Speed*(alert_vector_length)/3600.0; // the distance target travels in alert_vector_length seconds
 float projTargetLatR=asin(sin(targetLatR)*cos(targetDist/R)+cos(targetLatR)*sin(targetDist/R)*cos(targetTrackR)); // projected latitude
 float projTargetLonR=targetLonR+atan2(sin(targetTrackR)*sin(targetDist/R)*cos(targetLatR),cos(targetDist/R)-sin(targetLatR)*sin(projTargetLatR)); // projected longitude
 float projTargetLat=projTargetLatR*ToDegrees;
 float projTargetLon=remainder(projTargetLonR*ToDegrees+540, 360)-180;
 //calculate projected Ownship position in projSeconds;
 float ownshipLatR=ownShip.Lat*ToRad;
 float ownshipLonR=ownShip.Lon*ToRad;
 float ownshipTrackR=ownShip.Track*ToRad;
 float ownshipDist=ownShip.Speed*(alert_vector_length)/3600.0; // the distance target travels in projSeconds
 float projOwnshipLatR=asin(sin(ownshipLatR)*cos(ownshipDist/R)+cos(ownshipLatR)*sin(ownshipDist/R)*cos(ownshipTrackR)); // projected latitude
 float projOwnshipLonR=ownshipLonR+atan2(sin(ownshipTrackR)*sin(ownshipDist/R)*cos(ownshipLatR),cos(ownshipDist/R)-sin(ownshipLatR)*sin(projOwnshipLatR)); // projected longitude
 float projOwnshipLat=projOwnshipLatR*ToDegrees;
 float projOwnshipLon=remainder(projOwnshipLonR*ToDegrees+540, 360)-180;
 // get distance between original and projected positions
 if (velocity_vectors_intersect(myIntruder.Lat,myIntruder.Lon,projTargetLat,projTargetLon,ownShip.Lat,ownShip.Lon,projOwnshipLat,projOwnshipLon))
    {
    #ifdef DEBUG
    Serial.println();
    Serial.print("ALERT: ");
    Serial.print(myIntruder.Address,HEX);
    Serial.print(" ");
    Serial.print(" Intersecting Ownship within less than ");
    Serial.print(alert_vector_length);
    Serial.print(" seconds");    
    #endif
    return true;
    }
 return false;
}



bool velocity_vectors_intersect(float p0_x,float p0_y,float p1_x,float p1_y,float p2_x,float p2_y,float p3_x,float p3_y)
{
    float s1_x, s1_y, s2_x, s2_y;
    s1_x = p1_x - p0_x;
    s1_y = p1_y - p0_y;
    s2_x = p3_x - p2_x;
    s2_y = p3_y - p2_y;

    float s, t;
    float sdenom=(-s2_x * s1_y + s1_x * s2_y);
    if (sdenom==0) return false;
    s = (-s1_y * (p0_x - p2_x) + s1_x * (p0_y - p2_y)) / sdenom;
    float tdenom=(-s2_x * s1_y + s1_x * s2_y);
    if (tdenom==0) return false;
    t = ( s2_x * (p0_y - p2_y) - s2_y * (p0_x - p2_x)) / tdenom;

    if (s >= 0 && s <= 1 && t >= 0 && t <= 1)
    {
    return true;
    }
    return false;
}

float secondsToClose(intruder myIntruder)
    {
    float closing_speed=closingSpeed(myIntruder);
    float closing_time=myIntruder.Dist/closing_speed*3600; // time in second
    #ifdef DEBUG
    Serial.println();
    Serial.print(myIntruder.Address,HEX);
    Serial.print(" :");
    Serial.print(" ClosingSpeed: ");
    Serial.print(closing_speed);
    Serial.print("kts, ");
    Serial.print(" Distance: ");
    Serial.print(myIntruder.Dist);
    Serial.print("nm, ");
    Serial.print(", ClosingSeconds: ");
    Serial.print(closing_time);
    Serial.print("s, ");
    #endif  
    return closing_time;
    }

bool relAltitudeAlert(intruder myIntruder,float secondsToClose)
    {
    // calculate projected altitudes [alert_seconds_before] seconds into the future
    int32_t projOwnshipAlt=ownShip.Alt+ownShip.VertSpeed * secondsToClose / 60;    
    int32_t projIntruderAlt=myIntruder.Alt+myIntruder.VertSpeed * secondsToClose / 60;
    // calculate projected relative altitude
    int32_t projRelAlt=projIntruderAlt-projOwnshipAlt;
    // alert if the projected relative altitude is within our [alert_alt_diff] variable
    Serial.print(" projOwnShipAlt: ");
    Serial.print(projOwnshipAlt);
    Serial.print(" projIntruderAlt: ");
    Serial.print(projIntruderAlt);
    Serial.print(" projRelAlt: ");
    Serial.print(abs(projRelAlt));
    Serial.print(", ");
    if (abs(myIntruder.RelAlt)<=alert_alt_diff) return true; else
                                                                 if (abs(projRelAlt)<=alert_fuzzy_alt_diff) return true; else return false;
    }


bool ghostFilter(intruder myIntruder)
     {
     // ghost filter only on trackfile IDs
     if (isTrackFileID(myIntruder.AddressType) && myIntruder.Dist<=ghost_dist_diff && abs(myIntruder.RelAlt)<=ghost_alt_diff && (abs((360-myIntruder.Bearing) - (360-ownShip.Bearing))) <=ghost_track_diff )
     {
     #ifdef DEBUG
     Serial.println();
     Serial.print("Ghost filtered: ");
     Serial.print(myIntruder.Address);
     Serial.print(":");
     Serial.print(myIntruder.AddressType);
     Serial.write(", distance: ");
     Serial.print(myIntruder.Dist);  
     Serial.write(", track: ");
     Serial.print(myIntruder.Track);
     Serial.write(", rel alt: ");
     Serial.print(myIntruder.RelAlt);
     Serial.println();
     #endif
     return false;
     }
     else return true;
     }

     
float remainder(float a, int b)
{
 if (b==0) return 0;
  float fraction=a-(int)a;
  return ((int)a % b)+ fraction;

}

void prepareTrafficOut()
{
    if (callsignType <3) callsignType++; else callsignType=0; // cycle callsign types
    #ifdef DEBUG
    Serial.println("expiring intruders");
    #endif
    ExpireIntruders();
    coastTraffic();
    #ifdef DEBUG
    printTraffic();
    #endif
    
    // reset buffer fill index
    tempBufferFillIndex=0;
    
    //send heartbeat to bufferout
    // override Maintenance bit (turn it off)
    if (maintenanceOverride) bitWrite(heartbeatMessage[1],6,0);       
    addMessageToBuffer(heartbeatMessage,7);    
    // send ownship
    addMessageToBuffer(ownshipMessage,28);
    

    //send geoalt
    addMessageToBuffer(geoaltMessage,5);
    
    // send traffictargets
    if (totalIntruders<=maxIntrudersOut) totalIntrudersOut=totalIntruders; else totalIntrudersOut=maxIntrudersOut;
      
    for (int i=0; i<totalIntrudersOut;i++)
      {
       byte trafficmessage[28];
       IntruderToTrafficMessage(intruders[i],trafficmessage);
       addMessageToBuffer(trafficmessage,28);       
      }
    // send uplink messages    
    for (int i=0; i<uplinkMessageCount;i++)
        {
        addMessageToBuffer(uplinkMessageList[i].message,436);
        }
    #ifdef DEBUG
    Serial.println();
    Serial.print("Added ");Serial.print(uplinkMessageCount);Serial.println(" Uplinks");
    #endif

    // copy temp buffer to outputbuffer
    outputBufferLength=tempBufferFillIndex;
    for (int i=0;i<outputBufferLength;i++) outputBuffer[i]=tempBuffer[i];

    outputIndex=0;  // reset the output buffer index
    outputFlow=true; // let it go out the port
    //printMessage(outputBuffer,outputBufferLength,"OutputBUFFER: ");
}

void addMessageToBuffer(byte outputmessage[],int messagesize)
{
#ifdef SDLOGGING
int startBufferFillIndex=tempBufferFillIndex;
#endif
if (messagesize+tempBufferFillIndex>outputBufferSize)
    {
    #ifdef DEBUG
    Serial.println("Output Buffer Overflow");
    #endif
    return; // buffer overflow, do not add any more messages    
    }
// add message start marker
tempBuffer[tempBufferFillIndex]=0x7E;
tempBufferFillIndex++;

for (int i=0; i<messagesize;i++)
        {        
        if (outputmessage[i]==0x7D || outputmessage[i]== 0x7E)
          {
          // add escape character
          tempBuffer[tempBufferFillIndex]=0x7D;
          tempBufferFillIndex++;
          tempBuffer[tempBufferFillIndex]=outputmessage[i] ^ 0x20;
          } else  tempBuffer[tempBufferFillIndex]=outputmessage[i];
          
        tempBufferFillIndex++;
        }
// add two bytes of CRC;
uint16_t newCRC=crcCompute(outputmessage,messagesize);
tempBuffer[tempBufferFillIndex]=lowByte(newCRC);
tempBufferFillIndex++;
tempBuffer[tempBufferFillIndex]=highByte(newCRC);
tempBufferFillIndex++;
        
// add message end marker        
tempBuffer[tempBufferFillIndex]=0x7E;
tempBufferFillIndex++;

// sd card logging of GDL90output
  #ifdef SDLOGGING
      // output logging in hex format with timestamp
      if (sdLogging && outLogBufferIndex<80000)
        {
        // add timestamp
        char timestamp[12];
        sprintf (timestamp, "%012ld:", millis());
        memcpy(outLogBuffer+outLogBufferIndex,timestamp,13);
        outLogBufferIndex+=13;
        for (int i=startBufferFillIndex;i<=tempBufferFillIndex-1;i++)
            {
            char hexChar[3];
            sprintf(hexChar,"%02X ",tempBuffer[i]);
            memcpy(outLogBuffer+outLogBufferIndex,hexChar,3);
            outLogBufferIndex+=3;
            }           
        // addnew line
         outLogBuffer[outLogBufferIndex]=0x0A;
         outLogBufferIndex++;                            
        }
// output logging in binary GDL90 format, no timestamp
//       if (sdLogging && outLogBufferIndex<80000)
//        {
//        // add timestamp        
//        for (int i=startBufferFillIndex;i<=tempBufferFillIndex-1;i++)
//            {
//            outLogBuffer[outLogBufferIndex]=tempBuffer[i];
//            outLogBufferIndex++;
//            }
//        } 

  #endif
}

void IntruderToTrafficMessage(intruder theintruder, byte trafficmessage[])
{


// 1 byte messageid
trafficmessage[0]=0x14;
// 1 byte alert status + address type

#ifdef ALERTS
float closingseconds=secondsToClose(theintruder);
// set Alert flag if necessary
if (((closingseconds>=0 && closingseconds<=alert_seconds_before) || (vectorsIntersect(theintruder)) || theintruder.Dist<=alert_dist_diff || (closingseconds>=0 && closingseconds<=positive_closure_max_time && theintruder.Dist<=positive_closure_distance)) && relAltitudeAlert(theintruder,closingseconds))
    {
    theintruder.Alert=0x01;
    theintruder.AlertTime=millis();
    } else
          {
          if (theintruder.Alert==0x01 && (millis()-theintruder.AlertTime)> alert_hold_time*1000) theintruder.Alert=0x00; // reset alert if hold time has passed
          }
#endif          

#ifdef FRIENDS
bool Friend=isFriend(theintruder.Address);
if (Friend) theintruder.Alert=0x00;
#endif


trafficmessage[1]=theintruder.Alert<<4 | theintruder.AddressType;
// 3 bytes address
trafficmessage[2]=theintruder.Address >> 16 & 0xff;
trafficmessage[3]=theintruder.Address >> 8 & 0xff;
trafficmessage[4]=theintruder.Address & 0xff;
// 3 bytes latitude
long lat=theintruder.Lat/(180/pow(2,23));
trafficmessage[5]=lat >> 16 & 0xff;
trafficmessage[6]=lat >> 8 & 0xff;
trafficmessage[7]=lat & 0xff;
// 3 bytes longitude
long lon=theintruder.Lon/(180/pow(2,23));
trafficmessage[8]=lon >> 16 & 0xff;
trafficmessage[9]=lon >> 8 & 0xff;
trafficmessage[10]=lon & 0xff;
//altitude + misc (12 bits +4)
int16_t alt=(int)(theintruder.Alt+1000)/25;
trafficmessage[11]=byte((alt & 0xFF0) >> 4);
trafficmessage[12]=byte((alt & 0x00F) << 4) | (theintruder.Misc & 0x0F);


//FRIEND_dist_start
int8_t Nic=theintruder.Nic;
int8_t NacP=theintruder.NacP;

#ifdef FRIENDS
// if friend change integrity;
if (Friend && theintruder.Dist>=FRIEND_dist_start &&  theintruder.Dist<=FRIEND_dist_max)
    {
    Nic=1;
    NacP=1; 
    }
#endif
// 1 bytes nic +nacp
trafficmessage[13]=byte(Nic<<4 | (NacP & 0x0F));
// 3 bytes horizontal velocity+ vertical velocity
trafficmessage[14]=byte((theintruder.Speed & 0xFF0) >> 4);
int vspeed=theintruder.VertSpeed/64;
trafficmessage[15]=byte((theintruder.Speed & 0x00F) << 4) | byte((vspeed & 0x0F00)>>8);
trafficmessage[16]=byte(vspeed & 0xFF);
// 1 byte track
uint8_t track=theintruder.Track/1.3671875;
trafficmessage[17]=byte(track);
// 1 byte emitter category
trafficmessage[18]=theintruder.EmitterCategory;
// 8 bytes of callsign


#ifdef PLANEINFO
planeInfo thisPlane=getPlaneInfo(theintruder.Address);
if (thisPlane.valid) {
                      for (int k=0;k<8;k++) trafficmessage[k+19]=thisPlane.Model[k]; // send model info instead of N-number
                     }
                        else
                        {
                        for (int k=19;k<27;k++) trafficmessage[k]=theintruder.Name[k-19]; // if nothing in the database just display the original callsign
                        }
//#ifdef DEBUG                        
Serial.print("Callsign:");Serial.print(callsignType);Serial.print(":");for (int k=0;k<8;k++) Serial.print(char(trafficmessage[k+19])); Serial.println();
//#endif

#else
for (int k=19;k<27;k++) trafficmessage[k]=theintruder.Name[k-19]; // display the original callsign if no SD card.
#endif


// prioritycode and spare
trafficmessage[27]=theintruder.priorityCodeSpare;
}


void sendGdl90Bytes()
{
// called every few milliseconds, depending on bandwidth
if (outputFlow==true)
  {
  Serial2.write(outputBuffer[outputIndex]);
  outputIndex++;
  if (outputIndex==outputBufferLength)
          {
          outputFlow=false;
          }
  }
}

bool isTrackFileID(byte AddressType)
{
//if (Address >= 0x2A0001 && Address <= 0x3FFFFF) return true; else return false;
if (AddressType==3) return true; else return false;
}

#ifdef FRIENDS
bool isFriend(int32_t Address)
{
  for (int i=0;i<totalFriends;i++)
              {
              if (friendArray[i]==Address)
                      {
                      return true;                      
                      break; // exit loop if we found a friend                      
                      }                      
              }
  return false;               
}
#endif

#ifdef SDLOGGING
void saveLogBuffer()
{
if (sdLogging && inLogBufferIndex>=8192)
    {        
                        inLogFile = SD.open(filenameInLog, FILE_WRITE);
                              if (inLogFile && inLogBufferIndex>0)
                                {                                         
                                for (int i=0;i<inLogBufferIndex-1;i++)
                                    {
                                    inLogFile.print(inLogBuffer[i]);                                                                 
                                    }
                                inLogBufferIndex=0;    
                                inLogFile.close();
                                }

    
    }
  if (sdLogging && outLogBufferIndex>=4096)
    {        
                        outLogFile = SD.open(filenameOutLog, FILE_WRITE);
                              if (outLogFile && outLogBufferIndex>0)
                                {                                         
                                for (int i=0;i<outLogBufferIndex-1;i++)
                                    {
                                    outLogFile.write(outLogBuffer[i]);                                                                 
                                    }
                                outLogBufferIndex=0;    
                                outLogFile.close();
                                }

    
    }  
}
#endif

#ifdef PLANEINFO
planeInfo getPlaneInfo(int32_t Address)
{
planeInfo data;
char filename[13];
char readBuffer[57];
int bufferIndex=0;
int arrayIndex;

// check array first
arrayIndex=checkPlaneInfoArray(Address);
if (arrayIndex>=0)
    {    
    return planeInfoArray[arrayIndex];
    }

// not in array, check if Address should be on SD

data.Address=Address;
// space pad results
for (int i=0;i<8;i++)
    {
    data.Nnumber[i]=0x20;
    data.Manufacturer[i]=0x20;
    data.Model[i]=0x20;  
    }

if ((Address>>16 & 0xFF)<=0xA0 || (Address>>16 & 0xFF)>=0xAD)
  {
    data.valid=false;
    storePlaneInfo(data);
    #ifdef DEBUG
    Serial.println("not in SD database");
    #endif
    return data;
  }

    
// build file name
sprintf(filename,"/%02X/%02X/%02X.txt",(int)Address>>16 & 0x00FF,(int)Address>>8 & 0x00FF,(int)Address & 0x00FF);
#ifdef DEBUG
Serial.print("openfile:");Serial.println(filename);
#endif

if (SD.exists(filename))
{ 
PlaneInfoFile = SD.open(filename, FILE_READ);
       if (PlaneInfoFile) {
            while (PlaneInfoFile.available() && bufferIndex<=57) {
                                         readBuffer[bufferIndex]=PlaneInfoFile.read();
                                         bufferIndex++;
                                         }
            PlaneInfoFile.close();
            int commaCount=0;
            int arrayIndex=0;
            for (int i=0;i<bufferIndex;i++)
              {
              char c= readBuffer[i];
              if (c==',')
                    {
                    commaCount++;
                    arrayIndex=0;
                    } else
                          {
                          if (arrayIndex<8)
                                        {
                                        switch (commaCount) {
                                                            case 0:
                                                              if (arrayIndex==0)
                                                                  {
                                                                  data.Nnumber[arrayIndex]='N';
                                                                  arrayIndex++;
                                                                  }
                                                              data.Nnumber[arrayIndex]=c; arrayIndex++;
                                                              break;
                                                            case 1:
                                                              data.Manufacturer[arrayIndex]=c; arrayIndex++;
                                                              break;
                                                            case 2:
                                                              data.Model[arrayIndex]=c; arrayIndex++;                                                             
                                                              break;
                                                             }
                                        }                    
                          }
              }        
            data.valid=true;
             #ifdef DEBUG
             Serial.println("Got valid data from SD");
             #endif
             } else
                {
                data.valid=false;
                #ifdef DEBUG
                Serial.println("SD read Error");
                #endif
                }
} else
    {
     // file doesn't exist
    data.valid=false;
    #ifdef DEBUG
    Serial.println("Missing file");
    #endif
    }
// store planeInfo in array
storePlaneInfo(data);
return data;
}

int checkPlaneInfoArray(int32_t Address)
{
for (int i=0;i<maxPlaneInfoElements;i++)
    {
    if (planeInfoArray[i].Address==Address)
          {
          planeInfoArray[i].LastAccessTime=millis();
          return i;
          }
    } 
// return a 0 address result, not found in array
return -1;
}

void storePlaneInfo(planeInfo data)
{
int oldestIndex=0;
unsigned long oldestTime=millis();
data.LastAccessTime=millis();
// find an empty spot or the planeinfo accessed the most time ago

for (int i=0;i<maxPlaneInfoElements;i++)
  {
  if (planeInfoArray[i].Address==0) {
                                    // found an empty spot, add it here
                                    planeInfoArray[i]=data;                                                                       
                                    return;
                                   }
  if (planeInfoArray[i].LastAccessTime<oldestTime)
          {
          oldestTime=planeInfoArray[i].LastAccessTime;
          oldestIndex=i;
          } 
          
  }
 // add the data in the oldest spot;
 planeInfoArray[oldestIndex]=data; 
}
#endif
